#ifndef MESSAGE_CLI_H
#define MESSAGE_CLI_H

#include <zephyr/bluetooth/mesh.h>
#include <bluetooth/mesh/model_types.h>

#ifdef __cplusplus
extern "C" {
#endif


/** Company ID of the Bluetooth Mesh Message model. */
#define MESSAGE_MODEL_COMPANY_ID CONFIG_BT_COMPANY_ID

/** Model ID of the Bluetooth Mesh Message model. */
#define MESSAGE_MODEL_ID 0x000A

/** Non-private message opcode. */
#define MESSAGE_MODEL_OP_MESSAGE BT_MESH_MODEL_OP_3(0x0A, MESSAGE_MODEL_COMPANY_ID)

/** Private message opcode. */
#define MESSAGE_MODEL_OP_PRIVATE_MESSAGE BT_MESH_MODEL_OP_3(0x0B, MESSAGE_MODEL_COMPANY_ID)

/** Message reply opcode. */
#define MESSAGE_MODEL_OP_MESSAGE_REPLY BT_MESH_MODEL_OP_3(0x0C, MESSAGE_MODEL_COMPANY_ID)

/** Status message opcode. */
#define MESSAGE_MODEL_OP_STATUS BT_MESH_MODEL_OP_3(0x0D, MESSAGE_MODEL_COMPANY_ID)

/** Status get message opcode. */
#define MESSAGE_MODEL_OP_STATUS_GET BT_MESH_MODEL_OP_3(0x0E, MESSAGE_MODEL_COMPANY_ID)

/** Buffer opcode. */
#define MESSAGE_MODEL_OP_BUFFER BT_MESH_MODEL_OP_3(0x0F, MESSAGE_MODEL_COMPANY_ID)


#define MESSAGE_MODEL_MSG_MESSAGE_LENGTH 90
#define MESSAGE_MODEL_MSG_MINLEN_MESSAGE 1
#define MESSAGE_MODEL_MSG_MAXLEN_MESSAGE (MESSAGE_MODEL_MSG_MESSAGE_LENGTH + 1) /* + \0 */
#define MESSAGE_MODEL_MSG_LEN_MESSAGE_REPLY 0
#define MESSAGE_MODEL_MSG_LEN_STATUS 1
#define MESSAGE_MODEL_MSG_LEN_STATUS_GET 0

#define MESSAGE_MODEL_BUFF_MINLEN 1
#define MESSAGE_MODEL_BUFF_MAXLEN 100 //Max 113


/** Bluetooth Mesh Message model status values. */
enum message_model_status {
	MESSAGE_MODEL_STATUS_NEW_DATA,
    MESSAGE_MODEL_STATUS_DONE,
    MESSAGE_MODEL_STATUS_SLEEP,
};

/* Forward declaration of the Bluetooth Mesh Message model context. */
struct message_model_cli;

/** @def MESSAGE_MODEL_CLI
 *
 * @brief Bluetooth Mesh message model composition data entry.
 *
 * @param[in] _message Pointer to a @ref message_model_cli instance.
 */
#define MESSAGE_MODEL_CLI(_message)                                         \
		BT_MESH_MODEL_VND_CB(MESSAGE_MODEL_COMPANY_ID,       				\
			MESSAGE_MODEL_ID,                      							\
			_message_model_cli_op, &(_message)->pub,                   		\
			BT_MESH_MODEL_USER_DATA(struct message_model_cli, _message),    \
			&_message_model_cli_cb)

/** Bluetooth Message model handlers. */
struct message_model_handlers {

	/** @brief Called after the node has been provisioned, or after all
	 * mesh data has been loaded from persistent storage.
	 *
	 * @param[in] cli Message model client instance that has been started.
	 */
	void (*const start)(struct message_model_cli *cli);

	/** @brief Handler for a status message.
	 *
	 * @param[in] cli Message model client instance that received the text message.
	 * @param[in] ctx Context of the incoming message.
	 * @param[in] status Status of a Message model client that published
	 * the message.
	 */
	void (*const status)(struct message_model_cli *cli, struct bt_mesh_msg_ctx *ctx, enum message_model_status status);

	/** @brief Handler for a non-private text message.
	 *
	 * @param[in] cli Message model client instance that received the text message.
	 * @param[in] ctx Context of the incoming message.
	 * @param[in] msg Pointer to a received text message terminated with
	 * a null character, '\0'.
	 */
	void (*const message)(struct message_model_cli *cli, struct bt_mesh_msg_ctx *ctx, const uint8_t *msg);

	/** @brief Handler for a buffer data.
	 *
	 * @param[in] cli Message model client instance that received the buffer.
	 * @param[in] ctx Context of the incoming message.
	 * @param[in] buff Pointer to the received data buffer
	 * @param[in] len Length of the received data buffer.
	 */
	void (*const buffer)(struct message_model_cli *cli, struct bt_mesh_msg_ctx *ctx, const uint8_t *buff, uint16_t len);

	/** @brief Handler for a private message.
	 *
	 * @param[in] cli Message model client that received the text message.
	 * @param[in] ctx Context of the incoming message.
	 * @param[in] msg Pointer to a received text message terminated with
	 * a null character, '\0'.
	 */
	void (*const private_message)(struct message_model_cli *cli, struct bt_mesh_msg_ctx *ctx, const uint8_t *msg);

	/** @brief Handler for a reply on a private message.
	 *
	 * @param[in] cli Message model client instance that received the reply.
	 * @param[in] ctx Context of the incoming message.
	 */
	void (*const message_reply)(struct message_model_cli *cli, struct bt_mesh_msg_ctx *ctx);
};

/**
 * Bluetooth Mesh message model context.
 */
struct message_model_cli {
	/** Access model pointer. */
	struct bt_mesh_model *model;
	/** Publish parameters. */
	struct bt_mesh_model_pub pub;
	/** Publication message. */
	struct net_buf_simple pub_msg;
	/** Publication message buffer. */
	uint8_t buf[BT_MESH_MODEL_BUF_LEN(MESSAGE_MODEL_OP_MESSAGE, MESSAGE_MODEL_MSG_MAXLEN_MESSAGE)];
	/** Handler function structure. */
	const struct message_model_handlers *handlers;
	/** Current Status value. */
	enum message_model_status status;
};

/** @brief Set the client status and publish the status to the mesh network.
 *
 * @param[in] cli   Message model instance to set status on.
 * @param[in] status Status to be published.
 *
 * @retval 0 Successfully set the status and sent the message.
 * @retval -EADDRNOTAVAIL Publishing is not configured.
 * @retval -EAGAIN The device has not been provisioned.
 */
int message_model_status_set(struct message_model_cli *cli, enum message_model_status status);

/** @brief Get current status value of a message model client.
 *
 * @param[in] cli Message model instance to send the message.
 * @param[in] addr Address of the Message model to get status value of.
 *
 * @retval 0 Successfully sent the message.
 * @retval -EINVAL The model is not bound to an application key.
 * @retval -EAGAIN The device has not been provisioned.
 */
int message_model_status_get(struct message_model_cli *cli, uint16_t addr);

/** @brief Send a text message.
 *
 * @param[in] cli Message model instance to send the message.
 * @param[in] msg Pointer to a text message to send. Must be terminated with a null character, '\0'.
 *
 * @retval 0 Successfully sent the message.
 * @retval -EADDRNOTAVAIL Publishing is not configured.
 * @retval -EAGAIN The device has not been provisioned.
 */
int message_model_message_send(struct message_model_cli *cli, const uint8_t *msg);

/** @brief Send a buffer with data.
 *
 * @param[in] cli Message model instance to send the message.
 * @param[in] buffer Pointer to a buffer with data to send.
 * @param[in] buff_len Length of the data buffer.
 *
 * @retval 0 Successfully sent the message.
 * @retval -EADDRNOTAVAIL Publishing is not configured.
 * @retval -EAGAIN The device has not been provisioned.
 */
int message_model_buffer_send(struct message_model_cli *cli, const uint8_t *buffer, int buff_len);

/** @brief Send a text message to a specified destination.
 *
 * @param[in] cli  Message model instance to send the message.
 * @param[in] addr Address of the Message model to send message to.
 * @param[in] msg  Pointer to a text message to send. Must be terminated with a null character, '\0'.
 *
 * @retval 0 Successfully sent the message.
 * @retval -EINVAL The model is not bound to an application key.
 * @retval -EAGAIN The device has not been provisioned.
 */
int message_model_private_message_send(struct message_model_cli *cli, uint16_t addr, const uint8_t *msg);

/** @cond INTERNAL_HIDDEN */
extern const struct bt_mesh_model_op _message_model_cli_op[];
extern const struct bt_mesh_model_cb _message_model_cli_cb;
/** @endcond */

#ifdef __cplusplus
}
#endif

#endif